/**********************************************************************
  Filename    : Camera Remote Bluetooth IoT
  Description : Control android app Camera Remote Bluetooth remotely.
                Change camera modes: photo or video.
                Change flash modes: torch or auto
                Record video.
                Take photo.
  Version     : ESP32 CAM AI Thinker and for TFT ST7789 240x240 Color Display            
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/10/30
**********************************************************************/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include "BluetoothSerial.h"
#include "string.h"

#include "esp_timer.h"
#include "img_converters.h"
#include "fb_gfx.h"
#include "driver/ledc.h"
#include "sdkconfig.h"
#include "esp_camera.h"

#include <Arduino.h>
#include <SPI.h>
#include <Adafruit_GFX.h>
//#include <Adafruit_ST7789.h>
#include "Arduino_ST7789_Comp.h" // Hardware-specific library for ST7789
#include <TJpg_Decoder.h>

#if defined(ARDUINO_ARCH_ESP32) && defined(CONFIG_ARDUHAL_ESP_LOG)
#include "esp32-hal-log.h"
#define TAG ""
#else
#include "esp_log.h"
static const char *TAG = "camera_httpd";
#endif

#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
int led_duty = 0;
bool isStreaming = false;
#ifdef CONFIG_LED_LEDC_LOW_SPEED_MODE
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_LOW_SPEED_MODE
#else
#define CONFIG_LED_LEDC_SPEED_MODE LEDC_HIGH_SPEED_MODE
#endif
#endif

#define CAMERA_MODEL_AI_THINKER // Has PSRAM
#include "camera_pins.h"


typedef struct
{
    size_t size;  //number of values used for filtering
    size_t index; //current value index
    size_t count; //value count
    int sum;
    int *values; //array to be filled with values
} ra_filter_t;


#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
void enable_led(bool en)
{ // Turn LED On or Off
    int duty = en ? led_duty : 0;
    if (en && isStreaming && (led_duty > CONFIG_LED_MAX_INTENSITY))
    {
        duty = CONFIG_LED_MAX_INTENSITY;
    }
    ledc_set_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL, duty);
    ledc_update_duty(CONFIG_LED_LEDC_SPEED_MODE, CONFIG_LED_LEDC_CHANNEL);
    ESP_LOGI(TAG, "Set LED intensity to %d", duty);
}
#endif


#define MCU_MODE 3    //1: ESP32 WROOM 32, 2: ESP32 WROVER, 3: ESP32 CAM
#define DISPLAY_COLOR_MODE 2      // 0: no display, 1: Monochrome (SD1306, black and white), 2: Color (ST7789)

// ESP32 WROOM 32: I2C Connection *************************************
// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
// The pins for I2C are defined by the Wire-library. 
// On an arduino UNO:       A4(SDA), A5(SCL)
// On an arduino MEGA 2560: 20(SDA), 21(SCL)
// On an arduino LEONARDO:   2(SDA),  3(SCL), ...

// On ESP32 : 21(SDA), 22(SCL) ****************************************
// I2C:
// Default hardware pins: SDA (21), SCL (22)
//#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
//#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32

// uncomment and replace with statement bellow for this display
//Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// SD1306 *************************************************************
//#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
//#define OLED_MOSI   23
//#define OLED_CLK   18
//#define OLED_DC    15     //19
//#define OLED_CS    5
//#define OLED_RESET 4
//      
// uncomment and replace with statement bellow for this display   
// Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS);

// ESP32 WROVER: SPI Connection *****************************************
// SPI: VSPI (software spi: serial peripheral interface)
// Default hardware pins: MOSI (23), CLK (18), CS (5)
// Ueer selective pins: DC, RESET
// ST7789 ***************************************************************
#define TFT_CS         -1      //10
#define TFT_RST        2      //4      //9 // Or set to -1 and connect to Arduino RESET pin
#define TFT_DC         13     //8
#define TFT_MOSI       14 //23     //11  // Data out
#define TFT_SCLK       15 //18     //13  // Clock out
//// Initialize Adafruit ST7789 TFT library
Arduino_ST7789 tft = Arduino_ST7789(TFT_DC, TFT_RST, TFT_MOSI, TFT_SCLK);
#define SCREEN_WIDTH  240 
#define SCREEN_HEIGHT 240 

// Hardware  SPI
//#define TFT_CS         -1      // set pin number for library calling, but do not connect to mcu for ST7789 display and SPI mode (may need to connet for other display) 
//#define TFT_RST        4      // any available pin
//#define TFT_DC         15     // any available pin
//#define TFT_MOSI       23     // Data out
//#define TFT_SCLK       18 

// For 1.14", 1.3", 1.54", 1.69", and 2.0" TFT with ST7789:
// Hardware SPI
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);
// Software SPI
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);
// replace this statement with other display for correct display and versions
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);

// Bluetooth Status
int CONNECTED = 0;

BluetoothSerial SerialBT;

String dataValues[15]; 
boolean needDisplayRefresh = false; 
#define PIN_LED_FLASH    4    // find one available gpio for flash, if no hardware flash is available.

// This next function will be called during decoding of the jpeg file to
// render each block to the TFT.  If you use a different TFT library
// you will need to adapt this function to suit.
bool tft_output(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t* bitmap)
{
   // Stop further decoding as image is running off bottom of screen
  if ( y >= tft.height() ) return 0;

  // In ILI9341 library this function clips the image block at TFT boundaries
  //tft.drawRGBBitmap(x, y, bitmap, w, h);
  tft.drawImageF(x, y, w, h, bitmap);

  // Return 1 to decode next block
  return 1;
}


void setup() {

  Serial.begin(115200);

//  tft.init(240, 240, SPI_MODE3);           // Init ST7789 240x240, SPI_MODE3
  tft.init(240, 240);           
//  tft.setSPISpeed(40000000);
//  Serial.println(F("Display Mode: ST7789"));
//  // landscape mode (1, 3), portrait mode (0, 2)
  //tft.setRotation(2);
  tft.fillScreen(BLACK); 
  delay(100);
  TJpgDec.setJpgScale(1);
  // The decoder must be given the exact name of the rendering function above
  TJpgDec.setCallback(tft_output);
  delay(100);
  writeText(40, 100, 2, WHITE, "Camera Remote");
  writeText(40, 130, 2, WHITE, "Bluetooth");
  delay(2500);

//  if (!SPIFFS.begin()) {
//    Serial.println("SPIFFS initialisation failed!");
//    SPIFFS.begin(true);// Formats SPIFFS - could lose data https://github.com/espressif/arduino-esp32/issues/638
//  } else {
//    Serial.println("SPIFFS initialized.");
//    delay(500);
//    listFileList();  
//    delay(500);
//    deleteAllFiles();
//    delay(500);
//    listFileList();    
//  }


  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);    //disable brownout detector   ds

//  SerialBT.begin("Camera Remote Bluetooth");    //Bluetooth device name
  SerialBT.begin("Bluetooth Camera", false);    //Bluetooth device name, slave mode (connect from android device) 
  SerialBT.register_callback(callback);
  
  Serial.println("\nThe camera started, now you can connect (paired) from android remote mode by bluetooth!");
  
  delay(1000);
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 1);    //enable brownout detector

  if (CONNECTED) {
    // no codes will be run here
  }
  
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;
  
  // if PSRAM IC present, init with UXGA resolution and higher JPEG quality
  //                      for larger pre-allocated frame buffer.
  if(psramFound()){
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }

//  config.frame_size = FRAMESIZE_QVGA;
//  config.jpeg_quality = 10;
//  config.fb_count = 1;

#if defined(CAMERA_MODEL_ESP_EYE)
  pinMode(13, INPUT_PULLUP);
  pinMode(14, INPUT_PULLUP);
#endif
#if defined(CAMERA_MODEL_AI_THINKER)
//  pinMode(2, OUTPUT);
//  pinMode(15, OUTPUT);
//  pinMode(13, OUTPUT);
//  pinMode(14, OUTPUT);
  pinMode(PIN_LED_FLASH, OUTPUT);
#endif

  // camera init
  esp_err_t err = esp_camera_init(&config);
  if (err != ESP_OK) {
    Serial.printf("Camera init failed with error 0x%x", err);
    return;
  } else {
    Serial.println("Camera initialized successfully.");
  }

  sensor_t * s = esp_camera_sensor_get();
  // initial sensors are flipped vertically and colors are a bit saturated
  if (s->id.PID == OV2640_PID) {
    //s->set_vflip(s, 1); // flip it back
    s->set_brightness(s, 1); // up the brightness just a bit
    s->set_saturation(s, 1); // bit of the saturation
  }
  // drop down frame size for higher initial frame rate
  s->set_framesize(s, FRAMESIZE_HQVGA);

#if defined(CAMERA_MODEL_M5STACK_WIDE) || defined(CAMERA_MODEL_M5STACK_ESP32CAM)
  s->set_vflip(s, 1);
  s->set_hmirror(s, 1);
#endif

}

void changeFrameSize(String dataString) {

  parseData(dataString);

  String frameSize = dataValues[1];
  framesize_t fs = FRAMESIZE_HQVGA;
  
  if (frameSize.equalsIgnoreCase("small")) {
    fs = FRAMESIZE_QQVGA;  //QQVGA(160x120), FRAMESIZE_QQVGA;
  } else if (frameSize.equalsIgnoreCase("medium")) {
    fs = FRAMESIZE_HQVGA;  //HQVGA(240x176), FRAMESIZE_QVGA;
  } else if (frameSize.equalsIgnoreCase("large")) {
    fs = FRAMESIZE_QVGA;  //QVGA(320x240), FRAMESIZE_HVGA;  //FRAMESIZE_VGA;
  }

  sensor_t *s = esp_camera_sensor_get();
  int res = 0;
  res = s->set_framesize(s, fs);

}

void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;

    // clear display
    needDisplayRefresh = true;
    delay(500);
    tft.fillScreen(BLACK); 
    delay(500);
    writeText(40, 100, 2, WHITE, "Connected.");
    writeText(40, 160, 1, WHITE, "Camera Remote Bluetooth");
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;

    // resize camera frame size
    delay(500);
    changeFrameSize("framesize:medium");
    delay(1000);
    needDisplayRefresh = true;
  }
}

String filelist;
long file_number = 0;

void listFileList() {
  filelist = "";
  fs::File root = SPIFFS.open("/");

  fs::File file = root.openNextFile();
  while (file) {
    String fileName = file.name();
    filelist = filelist + ":" + fileName;
    file = root.openNextFile();
  }
  Serial.println(filelist);    
}
void sendFileList() {
  filelist = "";
  fs::File root = SPIFFS.open("/");

  fs::File file = root.openNextFile();
  while (file) {
    String fileName = file.name();
    filelist = filelist + ":" + fileName;
    file = root.openNextFile();
  }
  Serial.println(filelist);  

  SerialBT.print("filelist:");
  SerialBT.print(filelist + ":");
   // send new line
  SerialBT.write(13);
  SerialBT.write(10);   
}
void getLastFileNumber()
{
  fs::File root = SPIFFS.open("/");

  fs::File file = root.openNextFile();
  while (file) {
    String fileName = file.name();
    Serial.println(fileName);
    int fromUnderscore = fileName.lastIndexOf('_') + 1;
    int untilDot = fileName.lastIndexOf('.');
    String fileId = fileName.substring(fromUnderscore, untilDot);
    Serial.println(fileId);
    file_number = max(file_number, fileId.toInt()); // replace filenumber if fileId is higher
    file = root.openNextFile();
  }
}
void deleteFile(String fileName) {
  int fromUnderscore = fileName.lastIndexOf('_') + 1;
  int untilDot = fileName.lastIndexOf('.');
  String fileId = fileName.substring(fromUnderscore, untilDot);
  Serial.println(fileId);
  Serial.println("image delete");
  SPIFFS.remove("/thumb_" + fileId + ".jpg");
  SPIFFS.remove("/image_" + fileId + ".jpg");
  
  SerialBT.print("filedelete:");
  SerialBT.print(fileId + ":");
   // send new line
  SerialBT.write(13);
  SerialBT.write(10); 

  delay(1000);
  sendFileList();
}
void deleteAllFiles() {
  fs::File root = SPIFFS.open("/");
  fs::File file = root.openNextFile();
  while (file) {
    String fileName = file.name();
    Serial.println(fileName);
    SPIFFS.remove("/" + fileName);
    file = root.openNextFile();
  }  
}

// display text
void writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString) {
  tft.setTextWrap(true);
  tft.setCursor(x, y);
  tft.setTextSize(textSize);
  tft.setTextColor(textColor);
  tft.println(textString);
}

// clears the data value array
void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}
void parseData(String dataString) {
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }
}


void changeFlashMode(String dataString) {
  parseData(dataString);

  String flashMode = dataValues[1];
  if (flashMode.equalsIgnoreCase("on")) {
#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
    enable_led(true);
#endif
#if defined(CAMERA_MODEL_AI_THINKER)
    digitalWrite(PIN_LED_FLASH, HIGH);
#endif
  } else if (flashMode.equalsIgnoreCase("off")) {
#ifdef CONFIG_LED_ILLUMINATOR_ENABLED
    enable_led(false);
#endif 
#if defined(CAMERA_MODEL_AI_THINKER)
    digitalWrite(PIN_LED_FLASH, LOW);
#endif   
  }
}

bool imageDataProcessing = false;

void processData(String dataString) {

  //dataString = datastring;
  if (!dataString.startsWith("imagepixels")) {
    Serial.println("processData: " + dataString);
  }

  // parse data string into values
  parseData(dataString);

  // process parsed data values
  // data name: dataValues[0], 'photo', 'video', 'iot', 'imagedata', 'imagefixels', 'received', , 'imagefixels_color'
//
//  String dataName = dataValues[0];
  

  dataString = "";
}

void processImageData(String header) {
  // image data header received: "imagedata:xxxxx:16bit" image data pixel count, and color mode (16 bit or 8 bit)
//
//  if (!header.startsWith("imagedata")) {
//    imageDataProcessing = false;
//    return;
//  }

  Serial.println(header);

  parseData(header);
}


void processAction() {


        if (CONNECTED) {

//          SerialBT.println(actionString);
//          SerialBT.write(13);
//          SerialBT.write(10);
      
//          ACTION_STATUS = 1;
    
//          Serial.println(actionString);          
        }

        delay(100);     

//
//    if (ACTION_STATUS == 1) {
//
////      if (photo && video && flash && action && camera && displayMode) {
////        ACTION_STATUS = 0;
////        delay(10);      
////      }
//    
//    }

}

// Jpeg decoding and display jpeg image to display
//byte* jpegData;
//void displayJpeg(int dataLength) {
//  //byte* jpegData;
//  jpegData = (byte*)malloc(dataLength * sizeof(byte));
//
//  for(int i = 0; i < dataLength; i++) {
//    jpegData[i] = image_data_receiving[i];
//  }
//
//  TJpgDec.drawJpg(24, 24, jpegData, dataLength * sizeof(byte));
//
//  free(jpegData);
//}

uint8_t* frameData;




void displayCamImage(void)
{
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;    
    camera_fb_t *fb = esp_camera_fb_get();

    //tft.drawEsp32camImage(0, 0, 240, 480, (uint8_t*)fb->buf);

    if (!fb) {
        //ESP_LOGE(TAG, "Frame buffer could not be acquired");
        Serial.println("Frame buffer could not be acquired");
        delay(200);
        return;
    }

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    uint16_t w = 0, h = 0;
    TJpgDec.getJpgSize(&w, &h, fb->buf, fb->len);
    Serial.print("- Width = "); Serial.print(fb->width); Serial.print(", height = "); Serial.println(fb->height);   
    TJpgDec.drawJpg(0, 0, _jpg_buf, _jpg_buf_len * sizeof(byte));

    esp_camera_fb_return(fb);

    if (needDisplayRefresh) {
      tft.fillRect(0, 176, 240, 64, BLACK);
      needDisplayRefresh = false;
    }
    writeText(40, 200, 2, WHITE, "Not connected.");
}

esp_err_t camera_capture(){
    camera_fb_t * fb = NULL;
    size_t _jpg_buf_len = 0;
    uint8_t *_jpg_buf = NULL;
  
    //capture a frame
    fb = esp_camera_fb_get();
    if (!fb) {
        //ESP_LOGE(TAG, "Frame buffer could not be acquired");
        Serial.println("Frame buffer could not be acquired");
        delay(200);
        return ESP_FAIL;
    }

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;

    //replace this with your own function
    //display_image(fb->width, fb->height, fb->pixformat, fb->buf, fb->len);
//    uint16_t w = 0, h = 0;
//    TJpgDec.getJpgSize(&w, &h, fb->buf, fb->len);
//    Serial.print("- Width = "); Serial.print(fb->width); Serial.print(", height = "); Serial.println(fb->height);   
//    TJpgDec.drawJpg(0, 0, _jpg_buf, _jpg_buf_len * sizeof(byte));

    //return the frame buffer back to be reused
    if (fb)
    {
        Serial.print("width: ");
        Serial.print(fb->width);
        Serial.print(", height: ");
        Serial.print(fb->height);
        Serial.print(", len: ");
        Serial.println(_jpg_buf_len);

        if (CONNECTED) {
          
          // send frame data header
          SerialBT.print("frame_data");
          SerialBT.print(":width:");
          SerialBT.print(fb->width);
          SerialBT.print(":height:");
          SerialBT.print(fb->height);
          SerialBT.print(":len:");
          SerialBT.println(_jpg_buf_len);

          // send frame data
          //frameData = frameData[_jpg_buf_len];
          //memcpy(frameData, _jpg_buf, _jpg_buf_len);  
          
          //frameData = (uint8_t*)malloc(_jpg_buf_len * sizeof(uint8_t));
          //memmove(frameData, _jpg_buf, _jpg_buf_len);  
          
          //delay(10);       
          //SerialBT.write(frameData, _jpg_buf_len);         
          SerialBT.write(_jpg_buf, _jpg_buf_len);         

          // send new line
          SerialBT.write(13);
          SerialBT.write(10);
        }

        //delay(10);
        esp_camera_fb_return(fb);
        fb = NULL;
        _jpg_buf = NULL;

        
    } else if (_jpg_buf)
    {
        free(_jpg_buf);
        _jpg_buf = NULL;
    }    


    return ESP_OK;
}


bool takingPicture = false;
void takePictureAndSend() {
  takingPicture = true;
  sensor_t *s = esp_camera_sensor_get();
  //framesize_t fs = s->status.framesize;
  int res = 0;
  res = s->set_framesize(s, FRAMESIZE_UXGA);

  delay(500);

  camera_fb_t * fb = NULL;
  size_t _jpg_buf_len = 0;
  uint8_t *_jpg_buf = NULL;

  // wait for the auto exposure sets
  for (int i = 0; i < 6; i++) {
    fb = esp_camera_fb_get();
    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
    delay(100);  
  }

  
  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");
      
    //send failed status message
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:0");
    SerialBT.print(":height:0");
    SerialBT.print(":len:0:");
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);  
     
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:");
    SerialBT.print(fb->width);
    SerialBT.print(":height:");
    SerialBT.print(fb->height);
    SerialBT.print(":len:");
    SerialBT.println(_jpg_buf_len);
         
    SerialBT.write(_jpg_buf, _jpg_buf_len);   
    // send new line
    SerialBT.write(13);
    SerialBT.write(10);  

    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
  }

  sensor_t * s3 = esp_camera_sensor_get();
  s3->set_framesize(s3, FRAMESIZE_HQVGA);  
    
  delay(1000);
  
  takingPicture = false;
}
void takePictureAndSave() {
  takingPicture = true;

  //debug
  listFileList();

  camera_fb_t * fb = NULL;
  size_t _jpg_buf_len = 0;
  uint8_t *_jpg_buf = NULL;
  
  // save frame data
  getLastFileNumber();
  file_number += 1;
  fb = esp_camera_fb_get(); 
   
  //save thumb
  sensor_t * s = esp_camera_sensor_get();
  s->set_framesize(s, FRAMESIZE_QQVGA);    
  char *thumb_filename = (char*)malloc(23 + sizeof(file_number));
  sprintf(thumb_filename, "/spiffs/thumb_%d.jpg", file_number);
  Serial.println("Opening file: ");
  FILE *thumbnail = fopen(thumb_filename, "w");
  if (thumbnail != NULL)  {
    size_t err = fwrite(fb->buf, 1, fb->len, thumbnail);
    Serial.printf("File saved: %s\n", thumb_filename);
  }  else  {
    Serial.println("Could not open file");
    takingPicture = false;
    fclose(thumbnail);
    SerialBT.print("picture_data");
    SerialBT.print(":width:0");
    SerialBT.print(":height:0");
    SerialBT.print(":len:0:");
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);      
    return;
  }
  fclose(thumbnail);

  // large size image 
  sensor_t *s2 = esp_camera_sensor_get();
  //framesize_t fs = s->status.framesize;
  int res = 0;
  res = s2->set_framesize(s2, FRAMESIZE_HD);

  //delay(500);

  // wait for the auto exposure sets
  for (int i = 0; i < 6; i++) {
    fb = esp_camera_fb_get();
    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
    delay(100);  
  }
  
  fb = esp_camera_fb_get();
  if (!fb) {
    //ESP_LOGE(TAG, "Frame buffer could not be acquired");
    Serial.println("Frame buffer could not be acquired");
      
    //send failed status message
    // send frame data header
    SerialBT.print("picture_data");
    SerialBT.print(":width:0");
    SerialBT.print(":height:0");
    SerialBT.print(":len:0:");
     // send new line
    SerialBT.write(13);
    SerialBT.write(10);  
     
  } else {

    _jpg_buf_len = fb->len;
    _jpg_buf = fb->buf;    
    
//    // send frame data header
//    SerialBT.print("picture_data");
//    SerialBT.print(":width:");
//    SerialBT.print(fb->width);
//    SerialBT.print(":height:");
//    SerialBT.print(fb->height);
//    SerialBT.print(":len:");
//    SerialBT.println(_jpg_buf_len);
//         
//    SerialBT.write(_jpg_buf, _jpg_buf_len);   
//    // send new line
//    SerialBT.write(13);
//    SerialBT.write(10);  

    char *full_filename = (char*)malloc(23 + sizeof(file_number));
    sprintf(full_filename, "/spiffs/image_%d.jpg", file_number);
    Serial.println("Opening file: ");
    FILE *fullres = fopen(full_filename, "w");
    if (fullres != NULL)  {
      size_t err = fwrite(fb->buf, 1, fb->len, fullres);
      Serial.printf("File saved: %s\n", full_filename);
    }  else  {
      Serial.println("Could not open file");
      takingPicture = false;
      fclose(fullres);
      SerialBT.print("picture_data");
      SerialBT.print(":width:0");
      SerialBT.print(":height:0");
      SerialBT.print(":len:0:");
       // send new line
      SerialBT.write(13);
      SerialBT.write(10);  
      return;      
    }
    fclose(fullres);

    esp_camera_fb_return(fb);
    if (fb) {
      fb = NULL;
      _jpg_buf = NULL;
    } else if (_jpg_buf) {
      free(_jpg_buf);
      _jpg_buf = NULL;      
    }
  }

  SerialBT.print("filesave:");
  SerialBT.print("image_" + String(file_number) + ".jpg:");
   // send new line
  SerialBT.write(13);
  SerialBT.write(10);  

  sensor_t * s3 = esp_camera_sensor_get();
  s3->set_framesize(s3, FRAMESIZE_HQVGA);  
    
  delay(500);
  
  takingPicture = false;
}


void loop() {

//  currentTime = millis();
//  if (modeChangedTime < 0) {
//    modeChangedTime = currentTime;
//  }

  if (CONNECTED) {


    while(SerialBT.available())
    {
      
      String received = SerialBT.readStringUntil('\n');
      Serial.println(received);
      if (received.startsWith("flash")) {
        changeFlashMode(received);
      }
      if (!takingPicture) {
        if (received.startsWith("framesize")) {
          changeFrameSize(received);
        } else if (received.startsWith("frameget")) {
          takingPicture = true;
          delay(50);
          // send large picture to remote
          takePictureAndSend(); 
          // save images to flash memory
          //takePictureAndSave();     
        }      
      } else {
        delay(100);
      }
    }

    if (!takingPicture) {
      esp_err_t result = camera_capture();
    }
    
    processAction();

  } else { //end of: if (CONNECTED)

    // not connected
//    writeText(100, 100, 2, ST77XX_WHITE, "Not Connected");
    displayCamImage();
   
  }

  delay(50);
}
